import html
import json

from aiohttp import web

from .const import debugHTML, hintsHTML, tabHTML


def getLastRequests(history, devices, devicesName):
    out = "<h2>Last SEMP-Requests from SHM</h2>"
    if len(history) == 0:
        out += "None<br>"
    else:
        out += "<table><thead><tr><th>Time</th>\n"
        for d in devices:
            out += f"<th>{d[11:23]}<br>{devicesName[d].deviceName}</th>\n"
        out += "<th>Request IP</th></thead>\n<tbody>"

        for row in reversed(history):
            out += f"<tr><td>{row.time}</td>"
            for d in devices:
                value = ""
                if row.deviceData and d in row.deviceData:
                    value = f"{row.deviceData[d].power} ({row.deviceData[d].status})"
                if row.cmdData and d in row.cmdData:
                    value = f"Request: Turn {row.cmdData[d]['status']}"
                out += "<td>" + value + "</td>"
            out += f"<td>{row.remote} {row.typ}</td>"
            out += "</tr>"
        out += "</tbody></table>"
    return out


def getDeviceTab(dev):
    name = html.escape(dev.deviceName)
    out = "<pre>"
    out += json.dumps(
        {
            "deviceId": dev.deviceId,
            "deviceName": html.escape(dev.deviceName),
            "deviceType": dev.deviceType,
            "deviceSerial": html.escape(dev.deviceSerial),
            "deviceVendor": html.escape(dev.deviceVendor),
            "power": dev.power,
            "status": dev.status,
            "optionalEnergy": dev.optionalEnergy,
            "interruptionsAllowed": dev.interruptionsAllowed,
            "Max Consumption": dev.deviceMaxConsumption,
        },
        indent=4,
    )
    out += "</pre>"
    return tabHTML.format(tabname="tab" + name, data=out, check="")


async def statusPageRenderer(request: web.Request, deviceInfo, history):
    devicesSet = set()
    devicesName = {}
    for dataTuple in history:
        if dataTuple.deviceData:
            for devId in dataTuple.deviceData.keys():
                devicesSet.add(devId)
                if devId not in devicesName:
                    devicesName[devId] = dataTuple.deviceData[devId]
    devices = list(devicesSet)
    out = debugHTML

    out += "<h2>Current Status</h2>"
    out += "Select the device for more information:<br>"
    out += """<div class="container"><nav role="tab-control"><ul>"""

    out += '<li><label for="overview">Overview</label></li>'
    for dev in deviceInfo.values():
        out += f'<li><label for="tab{html.escape(dev.deviceName)}">{html.escape(dev.deviceName)}</label></li>'

    out += """ </ul></nav>  """
    out += '<div role="tabs"><section>\n'

    out += tabHTML.format(
        check='checked="checked"',
        tabname="overview",
        data=getLastRequests(history, devices, devicesName),
    )
    for dev in deviceInfo.values():
        out += getDeviceTab(dev)

    out += "\n</section></div>\n"
    out += hintsHTML
    return web.Response(text=out, content_type="text/html")



"""Test pysma init."""
import asyncio
import logging
from unittest.mock import patch
import json

import aiohttp
import pytest
import base64
import time
from pysma.device_em import SMAspeedwireEM
from pysma.exceptions import (
    SmaAuthenticationException,
    SmaConnectionException,
    SmaReadException,
)
from aioresponses import aioresponses
from . import MOCK_DEVICE, MOCK_L10N, mock_aioresponse

_LOGGER = logging.getLogger(__name__)

class Test_SMAEM_class:
    """Test the SMA class."""

    async def looper(self, sma: SMAspeedwireEM) -> None:
        """ Send fake packets from SunnyHomeManager2 """
        with open("tests/testdata/SunnyHomeManager2.json", "r") as file:
            data = json.load(file)
            data = base64.b64decode(data["packet"])
        for _ in range(1_000_000_000):
            await asyncio.sleep(0.6)
            sma.datagram_received(data, ("192.0.2.1", 4711))

    async def test_json_timeout_error(self, mock_aioresponse: aioresponses) -> None:
         """ Basic Tests """
         sma = SMAspeedwireEM()
         task = asyncio.create_task(self.looper(sma))
         await sma.new_session()
         device_info = await sma.device_info()
         print(device_info)
         for key in ["manufacturer", "name", "serial", "sw_version", "type" ]:
            assert key in device_info 
            assert len(device_info[key]) > 0 if isinstance(device_info[key], str) else True
            assert device_info[key] > 0 if isinstance(device_info[key], int) else True
         sensors = await sma.get_sensors()
         assert len(sensors) >= 17
         await sma.read(sensors)
         task.cancel()

    async def test_debug(self) -> None:
         """ Checks the encoding for the debug-information. """
         sma = SMAspeedwireEM()
         with open("tests/testdata/SunnyHomeManager2.json", "r") as file:
            data = json.load(file)
            packet = base64.b64decode(data["packet"])
            sma.di.last_packet = packet
            sma.di.last_valid_packet = packet
            debug = await sma.get_debug()
            print(debug)
            assert debug["last_packet"] == debug["last_valid_packet"] == data["packet"]  




import logging
import re
from datetime import datetime, timedelta

_LOGGER = logging.getLogger(__name__)


class sempTimeframe:
    start: datetime
    stop: datetime
    minRunningTime: int
    maxRunningTime: int

    def __init__(
        self,
        start: datetime,
        stop: datetime,
        minRunningTime: timedelta,
        maxRunningTime: timedelta,
    ):
        self.start = start
        self.stop = stop
        self.minRunningTime = int(minRunningTime.total_seconds())
        self.maxRunningTime = int(maxRunningTime.total_seconds())
        assert self.minRunningTime >= 0
        assert self.minRunningTime <= self.maxRunningTime


class sempDevice:

    deviceId: str
    deviceName: str
    deviceType: str
    deviceSerial: str
    deviceVendor: str
    deviceMaxConsumption: int
    deviceMinConsumption: int
    timeframes: list[sempTimeframe] = []

    @staticmethod
    def possibleDeviceType():
        return [
            "Other",
            "AirConditioning",
            "Charger",
            "DishWasher",
            "Dryer",
            #                "ElectricVehicle",
            #  "EVCharger",
            "Freezer",
            "Fridge",
            "Heater",
            "HeatPump",
            "Motor",
            "Pump",
            "WashingMachine",
        ]

    def __init__(
        self,
        deviceId: str,
        deviceName: str,
        deviceType: str,
        deviceSerial: str,
        deviceVendor: str,
        maxConsumption: int,
        minConsumption: int | None = None,
        minOnTime: timedelta | None = None,
        minOffTime: timedelta | None = None,
    ):
        assert re.match(
            "F-[0-9]{8}-[0-9]{12}-00", deviceId
        ), "DeviceID does not meet the requirements "
        if not (
            len(deviceId) == 26
            and deviceId.startswith("F-11223344-")
            and deviceId.endswith("-00")
        ):
            _LOGGER.debug(
                f"DeviceID {deviceId} does not meet the recommendations: F-11223344-XXXXXXXXXXXX-00"
            )
        assert deviceType in self.possibleDeviceType()
        self.deviceId = deviceId
        self.deviceName = deviceName
        self.deviceType = deviceType
        self.deviceSerial = deviceSerial
        self.deviceVendor = deviceVendor
        self.interruptionsAllowed = False
        self.power = 0
        self.deviceMaxConsumption = maxConsumption
        self.deviceMinConsumption = (
            minConsumption if minConsumption is not None else maxConsumption
        )
        self.minOnTime = int(minOnTime.total_seconds()) if minOnTime else 120
        self.minOffTime = int(minOffTime.total_seconds()) if minOffTime else 120
        self.status = "off"
        self.optionalEnergy = False
        # is set automatically from the timeframes (true, if mintime != maxtime).
        # <OptionalEnergy>false</OptionalEnergy>

        self.emSignalsAccepted = False
        # is set automatically from the timeframes (false if no timeframe defined).
        # EMSignalsAccepted (xs:
        # Bool that indicates if the device is currently considering the control signals or recommendations
        # provided by the energy manager or if it is in a mode which ignores the signals or recommendations
        # The value should be set to “false” if there is no timeframe in the planning request sec-
        # tion for the device at the moment.

    def setPowerStatus(self, power: int, status: str):
        assert status in ["on", "off", "offline"], f"Unknown Status {status}"
        assert power >= 0
        self.status = status
        self.power = power

    def setInterruptable(self, allowinterupts: bool) -> None:
        self.interruptionsAllowed = allowinterupts

    def setTimeframes(self, timeframes: list[sempTimeframe]):
        self.timeframes = timeframes

    def addTimeframe(self, timeframe: sempTimeframe):
        assert timeframe.minRunningTime <= timeframe.maxRunningTime
        assert timeframe.minRunningTime > 0
        if timeframe.maxRunningTime == timeframe.minRunningTime:
            # workaround
            timeframe.minRunningTime -= 1
        self.timeframes.append(timeframe)

    def update(self):
        """Based on the timeframes set optionalEnergy and emSignalsAccepted"""
        # TODO sort timeframe
        # TODO Was passiert wenn emSignalsAccepted forced False + timeframes
        self.optionalEnergy = False
        for tf in self.timeframes:
            if tf.minRunningTime != tf.maxRunningTime:
                self.optionalEnergy = True
        self.emSignalsAccepted = len(self.timeframes) > 0



"""Sensor definitions for SMA WebConnect library for Python."""

from .const import Identifier
from .const_webconnect import (
    DEVICE_INFO,
    ENERGY_METER_VIA_INVERTER,
    GENERIC_SENSORS,
    JMESPATHS_TAG,
    OPTIMIZERS_VIA_INVERTER,
)
from .sensor import Sensor

# Status - Operation
#: Status of the device
status = Sensor(
    "6180_08214800", Identifier.status, path=JMESPATHS_TAG, l10n_translate=True
)
#: General operating status
operating_status_general = Sensor(
    "6180_08412800",
    Identifier.operating_status_genereal,
    path=JMESPATHS_TAG,
    l10n_translate=True,
    enabled=False,
)
#: Operating status
operating_status = Sensor(
    "6180_08412B00",
    Identifier.operating_status,
    path=JMESPATHS_TAG,
    l10n_translate=True,
    enabled=False,
)
# Status - Operation - Inverter
#: General operating status
inverter_condition = Sensor(
    "6180_08414C00",
    Identifier.inverter_condition,
    path=JMESPATHS_TAG,
    l10n_translate=True,
    enabled=False,
)
#: Inverter Condition
inverter_system_init = Sensor(
    "6800_08811F00",
    Identifier.inverter_system_init,
    path=JMESPATHS_TAG,
    l10n_translate=True,
    enabled=False,
)
#: Grid connection status
grid_connection_status = Sensor(
    "6180_0846A700",
    Identifier.grid_connection_status,
    path=JMESPATHS_TAG,
    l10n_translate=True,
    enabled=False,
)
#: Grid relay status
grid_relay_status = Sensor(
    "6180_08416400",
    Identifier.grid_relay_status,
    path=JMESPATHS_TAG,
    l10n_translate=True,
    enabled=False,
)

# DC side - DC measurements PV
#: Current power generated by the solar panels (A side)
pv_power_a = Sensor("6380_40251E00_0", Identifier.pv_power_a, unit="W")
#: Current power generated by the solar panels (B side)
pv_power_b = Sensor("6380_40251E00_1", Identifier.pv_power_b, unit="W")
#: Current power generated by the solar panels (C side)
pv_power_c = Sensor("6380_40251E00_2", Identifier.pv_power_c, unit="W", enabled=False)
#: Current voltage generated by the solar panels (A side)
pv_voltage_a = Sensor(
    "6380_40451F00_0", Identifier.pv_voltage_a, unit="V", factor=100, enabled=False
)
#: Current voltage generated by the solar panels (B side)
pv_voltage_b = Sensor(
    "6380_40451F00_1", Identifier.pv_voltage_b, unit="V", factor=100, enabled=False
)
#: Current voltage generated by the solar panels (C side)
pv_voltage_c = Sensor(
    "6380_40451F00_2", Identifier.pv_voltage_c, unit="V", factor=100, enabled=False
)
#: Current amperage generated by the solar panels (A side)
pv_current_a = Sensor("6380_40452100_0", Identifier.pv_current_a, unit="A", factor=1000)
#: Current amperage generated by the solar panels (B side)
pv_current_b = Sensor("6380_40452100_1", Identifier.pv_current_b, unit="A", factor=1000)
#: Current amperage generated by the solar panels (C side)
pv_current_c = Sensor(
    "6380_40452100_2", Identifier.pv_current_c, unit="A", factor=1000, enabled=False
)
#: Isolation Resistance
pv_isolation_resistance = Sensor(
    "6102_00254F00_0",
    Identifier.pv_isolation_resistance,
    unit="kOhm",
    factor=1000,
    enabled=False,
)

# DC Side - Insulation monitoring
#: Insulation residual current
insulation_residual_current = Sensor(
    "6102_40254E00", Identifier.insulation_residual_current, unit="mA", enabled=False
)

# AC Side - PV Power measurements
#: Current power generated by the solar panels
pv_power = Sensor("6100_0046C200", Identifier.pv_power, unit="W", enabled=False)

# AC Side - Grid measurements
#: Power supplied to the grid. grid_power = power_l1 + power_l2 + power_l3
grid_power = Sensor("6100_40263F00", Identifier.grid_power, unit="W")
#: Grid frequency
frequency = Sensor(
    "6100_00465700", Identifier.frequency, unit="Hz", factor=100, enabled=False
)

# AC Side - Grid measurements - Active power
#: Power for phase 1
power_l1 = Sensor("6100_40464000", Identifier.power_l1, unit="W", enabled=False)
#: Power for phase 2
power_l2 = Sensor("6100_40464100", Identifier.power_l2, unit="W", enabled=False)
#: Power for phase 3
power_l3 = Sensor("6100_40464200", Identifier.power_l3, unit="W", enabled=False)

# AC Side - Grid measurements - Reactive power
#: Total Reactive Power
grid_reactive_power = Sensor(
    "6100_40265F00", Identifier.grid_reactive_power, unit="var", enabled=False
)
#: Reactive Power for phase 1
grid_reactive_power_l1 = Sensor(
    "6100_40666000", Identifier.grid_reactive_power_l1, unit="var", enabled=False
)
#: Reactive Power for phase 2
grid_reactive_power_l2 = Sensor(
    "6100_40666100", Identifier.grid_reactive_power_l2, unit="var", enabled=False
)
#: Reactive Power for phase 3
grid_reactive_power_l3 = Sensor(
    "6100_40666200", Identifier.grid_reactive_power_l3, unit="var", enabled=False
)

# AC Side - Grid measurements - Apparent power
#: Total Apparent Power
grid_apparent_power = Sensor(
    "6100_40666700", Identifier.grid_apparent_power, unit="VA", enabled=False
)
#: Apparent Power for phase 1
grid_apparent_power_l1 = Sensor(
    "6100_40666800", Identifier.grid_apparent_power_l1, unit="VA", enabled=False
)
#: Apparent Power for phase 2
grid_apparent_power_l2 = Sensor(
    "6100_40666900", Identifier.grid_apparent_power_l2, unit="VA", enabled=False
)
#: Apparent Power for phase 3
grid_apparent_power_l3 = Sensor(
    "6100_40666A00", Identifier.grid_apparent_power_l3, unit="VA", enabled=False
)


# AC Side - Grid measurements - Power factor
#: Grid Power factor
grid_power_factor = Sensor(
    "6100_00665900", Identifier.grid_power_factor, unit="", factor=1000, enabled=False
)
#: Grid Power factor excitation
grid_power_factor_excitation = Sensor(
    "6180_08465A00",
    Identifier.grid_power_factor_excitation,
    path=JMESPATHS_TAG,
    l10n_translate=True,
    enabled=False,
)

# AC Side - Grid measurements - Phase Current
#: Current for phase 1
current_l1 = Sensor(
    "6100_40465300", Identifier.current_l1, unit="A", factor=1000, enabled=False
)
#: Current for phase 2
current_l2 = Sensor(
    "6100_40465400", Identifier.current_l2, unit="A", factor=1000, enabled=False
)
#: Current for phase 3
current_l3 = Sensor(
    "6100_40465500", Identifier.current_l3, unit="A", factor=1000, enabled=False
)
#: Total Current
current_total = Sensor("6100_00664F00", Identifier.current_total, unit="A", factor=1000)

# AC Side - Grid measurements - Phase voltage
#: Voltage for phase 1
voltage_l1 = Sensor(
    "6100_00464800", Identifier.voltage_l1, unit="V", factor=100, enabled=False
)
#: Voltage for phase 2
voltage_l2 = Sensor(
    "6100_00464900", Identifier.voltage_l2, unit="V", factor=100, enabled=False
)
#: Voltage for phase 3
voltage_l3 = Sensor(
    "6100_00464A00", Identifier.voltage_l3, unit="V", factor=100, enabled=False
)

# AC Side - Measured values - energy
#: Total power yield from a solar installation
total_yield = Sensor("6400_00260100", Identifier.total_yield, unit="kWh", factor=1000)
#: The solar plant's yield for today
daily_yield = Sensor("6400_00262200", Identifier.daily_yield, unit="Wh")

# AC Side - Measured values - Grid measurements
#: Power supplied to grid measured by energy meter
metering_power_supplied = Sensor(
    "6100_40463600", Identifier.metering_power_supplied, unit="W"
)
#: Power absorbed fromgrid measured by energy meter
metering_power_absorbed = Sensor(
    "6100_40463700", Identifier.metering_power_absorbed, unit="W"
)
#: Grid frequency measured by energy meter
metering_frequency = Sensor(
    "6100_00468100", Identifier.metering_frequency, unit="Hz", factor=100
)
#: Total power supplied to the grid measured by energy meter
metering_total_yield = Sensor(
    "6400_00462400", Identifier.metering_total_yield, unit="kWh", factor=1000
)
#: Total power from the grid measured by energy meter
metering_total_absorbed = Sensor(
    "6400_00462500", Identifier.metering_total_absorbed, unit="kWh", factor=1000
)

# AC Side - Measured values - Phase currents
#: Current for phase 1 measured by energy meter
metering_current_l1 = Sensor(
    "6100_40466500", Identifier.metering_current_l1, unit="A", factor=1000
)
#: Current for phase 2 measured by energy meter
metering_current_l2 = Sensor(
    "6100_40466600", Identifier.metering_current_l2, unit="A", factor=1000
)
#: Current for phase 3 measured by energy meter
metering_current_l3 = Sensor(
    "6100_40466B00", Identifier.metering_current_l3, unit="A", factor=1000
)

# AC Side - Measured values - Phase voltage
#: Voltage for phase 1 measured by energy meter
metering_voltage_l1 = Sensor(
    "6100_0046E500", Identifier.metering_voltage_l1, unit="V", factor=100, enabled=False
)
#: Voltage for phase 2 measured by energy meter
metering_voltage_l2 = Sensor(
    "6100_0046E600", Identifier.metering_voltage_l2, unit="V", factor=100, enabled=False
)
#: Voltage for phase 3 measured by energy meter
metering_voltage_l3 = Sensor(
    "6100_0046E700", Identifier.metering_voltage_l3, unit="V", factor=100, enabled=False
)

# AC Side - Electricity meter - Measured values - Active power feed-in
#: Active Power feed-in for phase 1 measured by energy meter
metering_active_power_feed_l1 = Sensor(
    "6100_0046E800", Identifier.metering_active_power_feed_l1, unit="W"
)
#: Active Power feed-in  for phase 2 measured by energy meter
metering_active_power_feed_l2 = Sensor(
    "6100_0046E900", Identifier.metering_active_power_feed_l2, unit="W"
)
#: Active Power feed-in  for phase 3 measured by energy meter
metering_active_power_feed_l3 = Sensor(
    "6100_0046EA00", Identifier.metering_active_power_feed_l3, unit="W"
)

# AC Side - Electricity meter - Measured values - Active power drawn
#: Active Power drawn for phase 1 measured by energy meter
metering_active_power_draw_l1 = Sensor(
    "6100_0046EB00", Identifier.metering_active_power_draw_l1, unit="W"
)
#: Active Power drawn for phase 2 measured by energy meter
metering_active_power_draw_l2 = Sensor(
    "6100_0046EC00", Identifier.metering_active_power_draw_l2, unit="W"
)
#: Active Power drawn for phase 3 measured by energy meter
metering_active_power_draw_l3 = Sensor(
    "6100_0046ED00", Identifier.metering_active_power_draw_l3, unit="W"
)
#: Current power consumption measured by energy meter
metering_current_consumption = Sensor(
    "6100_00543100", Identifier.metering_current_consumption, unit="W", enabled=False
)
#: Total power consumption measured by energy meter
metering_total_consumption = Sensor(
    "6400_00543A00",
    Identifier.metering_total_consumption,
    unit="kWh",
    factor=1000,
    enabled=False,
)

# AC Side - PV generation
#: Total kWh generated to date
pv_gen_meter = Sensor("6400_0046C300", Identifier.pv_gen_meter, unit="kWh", factor=1000)

# AC Side - Emergency power measurements (Secure Power Supply)
#: Secure power supply voltage
sps_voltage = Sensor(
    "6100_0046C600", Identifier.sps_voltage, unit="V", factor=100, enabled=False
)
#: Secure power supply current
sps_current = Sensor(
    "6100_0046C700", Identifier.sps_current, unit="A", factor=1000, enabled=False
)
#: Secure power supply power
sps_power = Sensor("6100_0046C800", Identifier.sps_power, unit="W", enabled=False)

# PV Inverter Optimizers
#: Serial number of optimizer
optimizer_serial = Sensor("6800_10852600", Identifier.optimizer_serial)
#: Power supplied by optimizer
optimizer_power = Sensor("6100_40652A00", Identifier.optimizer_power, unit="W")
#: Current supplied by optimizer
optimizer_current = Sensor(
    "6100_40652900", Identifier.optimizer_current, unit="A", factor=1000, enabled=False
)
#: Voltage supplied by optimizer
optimizer_voltage = Sensor(
    "6100_40652800", Identifier.optimizer_voltage, unit="V", factor=100, enabled=False
)
#: Temperature of optimizer
optimizer_temp = Sensor(
    "6100_40652B00", Identifier.optimizer_temp, unit="°C", factor=10, enabled=False
)


# Battery (inverter) - Battery (general parameters)
#: Total battery state of charge
battery_soc_total = Sensor("6100_00295A00", Identifier.battery_soc_total, unit="%")
#: State of charge battery A
battery_soc_a = Sensor(
    "6100_00498F00_0", Identifier.battery_soc_a, unit="%", enabled=False
)
#: State of charge battery B
battery_soc_b = Sensor(
    "6100_00498F00_1", Identifier.battery_soc_b, unit="%", enabled=False
)
#: State of charge battery C
battery_soc_c = Sensor(
    "6100_00498F00_2", Identifier.battery_soc_c, unit="%", enabled=False
)
#: Voltage battery A
battery_voltage_a = Sensor(
    "6100_00495C00_0", Identifier.battery_voltage_a, unit="V", factor=100, enabled=False
)
#: Voltage battery B
battery_voltage_b = Sensor(
    "6100_00495C00_1", Identifier.battery_voltage_b, unit="V", factor=100, enabled=False
)
#: Voltage battery C
battery_voltage_c = Sensor(
    "6100_00495C00_2", Identifier.battery_voltage_c, unit="V", factor=100, enabled=False
)
#: Current battery A
battery_current_a = Sensor(
    "6100_40495D00_0", Identifier.battery_current_a, unit="A", factor=1000
)
#: Current battery B
battery_current_b = Sensor(
    "6100_40495D00_1", Identifier.battery_current_b, unit="A", factor=1000
)
#: Current battery C
battery_current_c = Sensor(
    "6100_40495D00_2", Identifier.battery_current_c, unit="A", factor=1000
)
#: Temperature battery A
battery_temp_a = Sensor(
    "6100_40495B00_0", Identifier.battery_temp_a, unit="°C", factor=10
)
#: Temperature battery B
battery_temp_b = Sensor(
    "6100_40495B00_1", Identifier.battery_temp_b, unit="°C", factor=10
)
#: Temperature battery C
battery_temp_c = Sensor(
    "6100_40495B00_2", Identifier.battery_temp_c, unit="°C", factor=10
)
#: Battery status operating mode
battery_status_operating_mode = Sensor(
    "6180_08495E00",
    Identifier.battery_status_operating_mode,
    path=JMESPATHS_TAG,
    l10n_translate=True,
)

# Battery (inverter) - Diagnosis
#: Total battery capacity
battery_capacity_total = Sensor(
    "6100_00696E00", Identifier.battery_capacity_total, unit="%"
)
#: Capacity battery A
battery_capacity_a = Sensor(
    "6100_00499100_0", Identifier.battery_capacity_a, unit="%", enabled=False
)
#: Capacity battery B
battery_capacity_b = Sensor(
    "6100_00499100_1", Identifier.battery_capacity_b, unit="%", enabled=False
)
#: Capacity battery C
battery_capacity_c = Sensor(
    "6100_00499100_2", Identifier.battery_capacity_c, unit="%", enabled=False
)

# Battery (inverter) - Charge (voltage)
#: Charging voltage battery A
battery_charging_voltage_a = Sensor(
    "6102_00493500_0",
    Identifier.battery_charging_voltage_a,
    unit="V",
    factor=100,
    enabled=False,
)
#: Charging voltage battery B
battery_charging_voltage_b = Sensor(
    "6102_00493500_1",
    Identifier.battery_charging_voltage_b,
    unit="V",
    factor=100,
    enabled=False,
)
#: Charging voltage battery C
battery_charging_voltage_c = Sensor(
    "6102_00493500_2",
    Identifier.battery_charging_voltage_c,
    unit="V",
    factor=100,
    enabled=False,
)

# Battery (inverter) - Battery charge (power & energy)
#: Total charging power
battery_power_charge_total = Sensor(
    "6100_00496900", Identifier.battery_power_charge_total, unit="W"
)
#: Charging power battery A
battery_power_charge_a = Sensor(
    "6100_00499300_0", Identifier.battery_power_charge_a, unit="W", enabled=False
)
#: Charging power battery B
battery_power_charge_b = Sensor(
    "6100_00499300_1", Identifier.battery_power_charge_b, unit="W", enabled=False
)
#: Charging power battery C
battery_power_charge_c = Sensor(
    "6100_00499300_2", Identifier.battery_power_charge_c, unit="W", enabled=False
)
#: Total charge
battery_charge_total = Sensor(
    "6400_00496700", Identifier.battery_charge_total, unit="kWh", factor=1000
)
#: Charge battery A
battery_charge_a = Sensor(
    "6400_00499500_0",
    Identifier.battery_charge_a,
    unit="kWh",
    factor=1000,
    enabled=False,
)
#: Charge battery B
battery_charge_b = Sensor(
    "6400_00499500_1",
    Identifier.battery_charge_b,
    unit="kWh",
    factor=1000,
    enabled=False,
)
#: Charge battery C
battery_charge_c = Sensor(
    "6400_00499500_2",
    Identifier.battery_charge_c,
    unit="kWh",
    factor=1000,
    enabled=False,
)

# Battery (inverter) - Battery discharge (power & energy)
#: Total discharging power
battery_power_discharge_total = Sensor(
    "6100_00496A00", Identifier.battery_power_discharge_total, unit="W"
)
#: Disharging power battery A
battery_power_discharge_a = Sensor(
    "6100_00499400_0", Identifier.battery_power_discharge_a, unit="W", enabled=False
)
#: Disharging power battery B
battery_power_discharge_b = Sensor(
    "6100_00499400_1", Identifier.battery_power_discharge_b, unit="W", enabled=False
)
#: Disharging power battery C
battery_power_discharge_c = Sensor(
    "6100_00499400_2", Identifier.battery_power_discharge_c, unit="W", enabled=False
)
#: Total discharge
battery_discharge_total = Sensor(
    "6400_00496800", Identifier.battery_discharge_total, unit="kWh", factor=1000
)
#: Discharge battery A
battery_discharge_a = Sensor(
    "6400_00499600_0",
    Identifier.battery_discharge_a,
    unit="kWh",
    factor=1000,
    enabled=False,
)
#: Discharge battery B
battery_discharge_b = Sensor(
    "6400_00499600_1",
    Identifier.battery_discharge_b,
    unit="kWh",
    factor=1000,
    enabled=False,
)
#: Discharge battery C
battery_discharge_c = Sensor(
    "6400_00499600_2",
    Identifier.battery_discharge_c,
    unit="kWh",
    factor=1000,
    enabled=False,
)

# Device Parameters
# Type Label - Type Label
#: Device serial number
serial_number = Sensor("6800_00A21E00", Identifier.serial_number)
#: Device name
device_name = Sensor("6800_10821E00", Identifier.device_name)
#: Device type
device_type = Sensor(
    "6800_08822000", Identifier.device_type, path=JMESPATHS_TAG, l10n_translate=True
)
#: Device manufactorer
device_manufacturer = Sensor(
    "6800_08822B00",
    Identifier.device_manufacturer,
    path=JMESPATHS_TAG,
    l10n_translate=True,
)
#: Device software version
device_sw_version = Sensor("6800_00823400", Identifier.device_sw_version)

# Device - Inverter
#: Power limit of the Inverter
inverter_power_limit = Sensor(
    "6800_00832A00", Identifier.inverter_power_limit, unit="W"
)


active_power_limitation = Sensor(
    "6802_00832B00", Identifier.active_power_limitation, unit="W"
)


# System communication - Meter on Speedwire
#: Serial number of energy meter
energy_meter = Sensor("6800_008AA300", Identifier.energy_meter)


sensor_map = {
    GENERIC_SENSORS: [
        battery_capacity_a,
        battery_capacity_b,
        battery_capacity_c,
        battery_capacity_total,
        battery_charge_a,
        battery_charge_b,
        battery_charge_c,
        battery_charge_total,
        battery_charging_voltage_a,
        battery_charging_voltage_b,
        battery_charging_voltage_c,
        battery_current_a,
        battery_current_b,
        battery_current_c,
        battery_discharge_a,
        battery_discharge_b,
        battery_discharge_c,
        battery_discharge_total,
        battery_power_charge_a,
        battery_power_charge_b,
        battery_power_charge_c,
        battery_power_charge_total,
        battery_power_discharge_a,
        battery_power_discharge_b,
        battery_power_discharge_c,
        battery_power_discharge_total,
        battery_soc_a,
        battery_soc_b,
        battery_soc_c,
        battery_soc_total,
        battery_status_operating_mode,
        battery_temp_a,
        battery_temp_b,
        battery_temp_c,
        battery_voltage_a,
        battery_voltage_b,
        battery_voltage_c,
        current_l1,
        current_l2,
        current_l3,
        current_total,
        daily_yield,
        frequency,
        grid_apparent_power,
        grid_apparent_power_l1,
        grid_apparent_power_l2,
        grid_apparent_power_l3,
        grid_connection_status,
        grid_power,
        grid_power_factor,
        grid_power_factor_excitation,
        grid_reactive_power,
        grid_reactive_power_l1,
        grid_reactive_power_l2,
        grid_reactive_power_l3,
        grid_relay_status,
        insulation_residual_current,
        inverter_condition,
        inverter_power_limit,
        active_power_limitation,
        inverter_system_init,
        operating_status_general,
        operating_status,
        power_l1,
        power_l2,
        power_l3,
        pv_current_a,
        pv_current_b,
        pv_current_c,
        pv_gen_meter,
        pv_isolation_resistance,
        pv_power,
        pv_power_a,
        pv_power_b,
        pv_power_c,
        pv_voltage_a,
        pv_voltage_b,
        pv_voltage_c,
        sps_current,
        sps_power,
        sps_voltage,
        status,
        total_yield,
        voltage_l1,
        voltage_l2,
        voltage_l3,
    ],
    OPTIMIZERS_VIA_INVERTER: [
        optimizer_power,
        optimizer_current,
        optimizer_voltage,
        optimizer_temp,
    ],
    ENERGY_METER_VIA_INVERTER: [
        metering_power_supplied,
        metering_power_absorbed,
        metering_frequency,
        metering_total_yield,
        metering_total_absorbed,
        metering_current_l1,
        metering_current_l2,
        metering_current_l3,
        metering_voltage_l1,
        metering_voltage_l2,
        metering_voltage_l3,
        metering_active_power_feed_l1,
        metering_active_power_feed_l2,
        metering_active_power_feed_l3,
        metering_active_power_draw_l1,
        metering_active_power_draw_l2,
        metering_active_power_draw_l3,
        metering_current_consumption,
        metering_total_consumption,
    ],
    DEVICE_INFO: [
        serial_number,
        device_name,
        device_type,
        device_manufacturer,
        device_sw_version,
    ],
}



"""Test pysma const file."""

import pysma.definitions_webconnect
from pysma.sensor import Sensor
from pysma.definitions_ennexos import ennexosSensorProfiles
from pysma.definitions_speedwire import commands

def test_duplicate_sensors_webconnect():
    """Test if defined sensors have unique key and name."""
    variables = vars(pysma.definitions_webconnect)
    found_keys = []
    found_names = []
    for value in variables.values():
        if isinstance(value, Sensor):
            found_key = f"{value.key}_{value.key_idx}"
            found_name = value.name

            assert found_key not in found_keys
            found_keys.append(found_key)

            assert found_name not in found_names
            found_names.append(found_name)


def test_sensor_map():
    """Test if all map entries only contain unique items."""
    for sensor_map in pysma.definitions_webconnect.sensor_map.values():
        unique_items = list({f"{s.key}_{s.key_idx}": s for s in sensor_map}.values())
        assert unique_items == sensor_map


def test_sameunit_for_all_sensors():
    webconnect_sensors = []
    for sensors in pysma.definitions_webconnect.sensor_map.values():
        for sensor in sensors:
            if not isinstance(sensor, Sensor):
                print(f"Not a Sesnsor {type(sensor)}: {sensor}")
            else:
                webconnect_sensors.append(sensor)

    enneox_sensors =  []
    for sensors in ennexosSensorProfiles.values():
        for sensor in sensors:
            if not isinstance(sensor, Sensor):
                print(f"Not a Sesnsor {type(sensor)}: {sensor}")
            else:
                if sensor.name:
                    enneox_sensors.append(sensor)

    speedwire_sensors = []
    for c in commands.values():
        for r in c.get("registers",[]):
            if "sensor" in r:
                sensor = r["sensor"]
                if not isinstance(sensor, Sensor):
                    print(f"Not a Sesnsor {type(sensor)}: {sensor}")
                else:
                    speedwire_sensors.append(sensor)

    id2unit = {}
    errCount = 0
    for arr in [webconnect_sensors, enneox_sensors, speedwire_sensors]:
        for sen in arr:
            if sen.name in id2unit:
                if sen.unit != id2unit[sen.name]:
                    errCount += 1
                    print(f"inconsistent units {sen.name} {sen.unit} <=>  {id2unit[sen.name]}")
            else:
                id2unit[sen.name] = sen.unit
    assert(errCount == 0)



"""
Implementation for SMA Speedwire

Improved with Information from https://github.com/mhop/fhem-mirror/blob/master/fhem/FHEM/76_SMAInverter.pm
Receiver classes completely reimplemented by little.yoda

"""

import asyncio
import binascii
import collections
import copy
import logging
import struct
import time
from asyncio import DatagramProtocol, Future
from typing import Any, Dict, List, Optional

from .const import SMATagList
from .definitions_speedwire import (
    SpeedwireFrame,
    commands,
    responseDef,
    speedwireHeader,
    speedwireHeader6065,
)
from .device import Device, DeviceInformation, DiscoveryInformation
from .exceptions import (
    SmaAuthenticationException,
    SmaConnectionException,
    SmaReadException,
)
from .helpers import version_int_to_string
from .sensor import Sensor, Sensors

_LOGGER = logging.getLogger(__name__)


class SMAClientProtocol(DatagramProtocol):
    """Basic Class for communication"""

    _commandFuture: Future[Any] | None = None

    debug: Dict[str, Any] = {
        "msg": collections.deque(maxlen=len(commands) * 10),
        "data": {},
        "unfinished": set(),
        "ids": set(),
        "sendcounter": 0,
        "resendcounter": 0,
        "failedCounter": 0,
    }

    def __init__(
        self, password: str, on_connection_lost: Future, options: Dict[str, Any]
    ):
        self._lastSend: float = 0
        self._firstSend: float | None = None
        self.speedwire = SpeedwireFrame()
        self._transport = None
        self.password = password
        self.on_connection_lost = on_connection_lost
        self.cmds: list[str] = []
        self.cmdidx = 0
        self.future: Future | None = None
        self.data_values: dict[str, Any] = {}
        self.sensors: dict[str, Sensor] = {}
        self._group = ""
        self._resendcounter = 0
        self._failedCounter = 0
        self._sendCounter = 0
        self._commandTimeout = float(options.get("commandTimeout", 0.5))
        self._commandDelay = float(options.get("commandDelay", 0.0))
        self._overallTimeout = float(
            options.get("overallTimeout", 5 + len(commands) * (self._commandDelay))
        )
        self.allCmds: list[str] = []
        self.allCmds.extend(commands.keys())
        self.allCmds.remove("login")
        self.allCmds.remove("logoff")

    def connection_made(self, transport: Any) -> None:
        self._transport = transport

    async def controller(self) -> None:
        try:
            if self._resendcounter == 0:
                self.debug["sendcounter"] += 1
                self._sendCounter += 1
            if self._commandFuture is None:
                raise RuntimeError("_commandFuture not send")
            await asyncio.wait_for(self._commandFuture, timeout=self._commandTimeout)
            self.cmdidx += 1
            self._resendcounter = 0
        except (asyncio.TimeoutError, RuntimeError):
            _LOGGER.debug(f"Timeout in command. Resendcounter: {self._resendcounter}")
            self._resendcounter += 1
            self.debug["resendcounter"] += 1
            if self._resendcounter > 2:
                # Giving up. Next Command
                _LOGGER.debug("Timeout in command")
                self.cmdidx += 1
                self._resendcounter = 0
                self._failedCounter += 1
                self.debug["failedCounter"] += 1
        await self._send_next_command()

    def _confirm_repsonse(self, code: int = -1):
        """Mark the commandFuture as done"""
        if self._commandFuture is None or self._commandFuture.done():
            _LOGGER.debug(f"unexpected message {code:08X}")
            return
        self._commandFuture.set_result(True)

    async def start_query(self, cmds: List, future: Future, group: str) -> None:
        self.cmds = ["login"]
        self.cmds.extend(cmds)
        self.future = future
        self.cmdidx = 0
        self._failedCounter = 0
        self._sendCounter = 0
        self._group = group
        self.data_values = {}
        self.sensors = {}
        _LOGGER.debug("Sending login")
        self.debug["msg"].append(["SEND", "login"])
        self._firstSend = time.time()
        await self._send_next_command()

    def connection_lost(self, exc: Exception | None) -> None:
        """connection lost handler"""
        _LOGGER.debug("Connection lost: %s %s", type(exc), exc)
        self.on_connection_lost.set_result(True)

    def _send_command(self, cmd: bytes) -> None:
        """Send the Command"""
        _LOGGER.debug(
            f"Sending command [{len(cmd)}] -- {binascii.hexlify(cmd).upper()}"  # type: ignore[str-bytes-safe]
        )
        self._commandFuture = asyncio.get_running_loop().create_future()
        asyncio.get_running_loop().create_task(self.controller())
        if self._transport is None:
            raise RuntimeError("Transport is None")
        self._transport.sendto(cmd)

    async def _send_next_command(self) -> None:
        """Send the next command in the list"""
        if not self.future:
            return
        if self.cmdidx >= len(self.cmds):
            # All commands send. Clean up.
            f = self.future
            self.future = None
            await asyncio.sleep(0.2)  # Wait for delayed responses
            self.debug["data"] = self.data_values
            self.cmds = []
            self.cmdidx = 0
            f.set_result(True)
            if self._firstSend:
                self.debug["msg"].append(
                    ["TOTAL", 0, "", round(time.time() - self._firstSend, 2)]
                )
                self._firstSend = None

        else:
            if self._resendcounter == 0:
                await asyncio.sleep(self._commandDelay)
            # Send the next command
            self.debug["msg"].append(["SEND", self.cmds[self.cmdidx]])
            _LOGGER.debug("Sending " + self.cmds[self.cmdidx])
            self._lastSend = time.time()
            if (self.cmds[self.cmdidx]) == "login":
                groupidx = ["user", "installer"].index(self._group) == 1
                self._send_command(
                    self.speedwire.getLoginFrame(self.password, 0, groupidx)
                )
            else:
                self._send_command(
                    self.speedwire.getQueryFrame(
                        self.password, 0, self.cmds[self.cmdidx]
                    )
                )

    def _getFormat(self, handler: dict) -> tuple:
        """Return the necessary information for extracting the information"""
        converter = None
        format = handler.get("format", "")
        if format == "int":
            format = "<l"
        elif format == "" or format == "uint":
            format = "<L"
        elif format == "version":
            format = "<L"
            converter = version_int_to_string
        else:
            raise ValueError(f"Unknown Format {format}")
        size = struct.calcsize(format)
        return (format, size, converter)

    def handle_login(self, msg: speedwireHeader6065) -> None:
        """Is called if a login response is received"""
        _LOGGER.debug("Login rppsonse received!")
        self.sensors = {}
        self.data_values = {"error": msg.error}
        self.data_values["serial"] = str(msg.src_serial)
        if msg.error == 256:
            _LOGGER.error("Login failed!")
            if self.future:
                self.future.set_exception(
                    SmaAuthenticationException(
                        "Login failed! Credentials wrong (user/install or password)"
                    )
                )

    def handle_newvalue(self, sensor: Sensor, value: Any, overwrite: bool) -> None:
        """Set the new value to the sensor"""
        if value is None:
            return
        sen = copy.copy(sensor)
        if sen.factor and sen.factor != 1:
            value /= sen.factor
        sen.value = value
        if sen.key in self.sensors:
            oldValue = self.sensors[sen.key].value
            if oldValue != value:
                # _LOGGER.warning(
                #     f"Sensors {sen.key} {sen.name} Old Value: {oldValue} New values: {sen.value} Overwrite: {overwrite}"
                # )
                if not overwrite:
                    value = oldValue
        self.sensors[sen.key] = sen
        self.data_values[sen.key] = value

    def extractvalues(self, handler: Dict, subdata: bytes) -> list[Any]:
        (formatdef, size, converter) = self._getFormat(handler)
        values = []
        for idx in range(8, len(subdata), size):
            v = struct.unpack(formatdef, subdata[idx : idx + size])[0]
            if v in [0xFFFFFFFF, 0x80000000, 0xFFFFFFEC, -0x80000000, 0xFFFFFE]:
                v = None
            else:
                if converter:
                    v = converter(v)
                if "mask" in handler:
                    v = v & handler["mask"]
            values.append(v)
        return values

    def fixID(self, orig: str) -> str:
        if orig in responseDef:
            return orig
        for code in responseDef.keys():
            if code[0:7] == orig[:7]:
                return code
        return orig

    def handle_register(self, subdata: bytes, register_idx: int) -> None:
        """Handle the payload with all the registers"""
        code = int.from_bytes(subdata[0:4], "little")
        # c = f"{(code & 0xFFFFFFFF):08X}"
        c = f"{code:08X}"
        msec = int.from_bytes(subdata[4:8], "little")  # noqa: F841

        # Fix for strange response codes
        self.debug["ids"].add(c[6:])
        self._id = c[6:]
        c = self.fixID(c)

        # Handle unknown Responses
        if c not in responseDef:
            values = []
            valuesPos = []
            for idx in range(8, len(subdata), 4):
                v = struct.unpack("<l", subdata[idx : idx + 4])[0]
                values.append(v)
                valuesPos.append(f"{idx + 54}")
            _LOGGER.warning(f"No Handler for {c}: {values} @ {valuesPos}")
            self.debug["unfinished"].add(f"{c}")
            return

        # Handle known repsones
        for handler in responseDef[c]:
            values = self.extractvalues(handler, subdata)
            if "sensor" not in handler:
                continue
            v = None
            if handler["idx"] == 0xFF:
                """For some responses, a list is returned and the correct value
                within this list is marked by the top 8 bits."""
                for origValue in values:
                    if origValue is not None and (origValue & 0xFF000000) > 0:
                        v = origValue & 0x00FFFFFF
                        break
            else:
                v = values[handler["idx"]]

            sensor = handler["sensor"]

            # Special handling for a response that returns two values under the same code
            if isinstance(sensor, List):
                if register_idx >= len(sensor):
                    _LOGGER.warning(
                        f"No Handler for {c} at register idx {register_idx}: {values}"
                    )
                    continue
                _LOGGER.debug(
                    f"Special Handler for {c} at register idx {register_idx}: {values}"
                )
                sensor = sensor[register_idx]
            _LOGGER.debug(
                f"ID: {self._id} Values {sensor.name}/{sensor.key}: {v} {values}"
            )
            self.handle_newvalue(sensor, v, handler.get("overwrite", True))

    # Unfortunately, there is no known method of determining the size of the registers
    # from the message. Therefore, the register size is determined from the number of
    # registers and the size of the payload.
    def calc_register(self, data: bytes, msg: speedwireHeader6065) -> tuple:
        cnt_registers = msg.lastRegister - msg.firstRegister + 1
        size_datapayload = len(data) - 54 - 4
        size_registers = (
            size_datapayload // cnt_registers
            if size_datapayload % cnt_registers == 0
            else -1
        )
        return (cnt_registers, size_registers)

    # Main routine for processing received messages.
    def datagram_received(self, data: bytes, addr: tuple[str, int]) -> None:
        _LOGGER.debug(f"RECV: {addr} Len:{len(data)} {binascii.hexlify(data).upper()}")  # type: ignore[str-bytes-safe]
        delta = 0.0
        if self._lastSend > 0:
            delta = time.time() - self._lastSend
            self._lastSend = 0
        self.debug["msg"].append(
            [
                "RECV",
                len(data),
                binascii.hexlify(data).upper().decode("utf-8"),
                round(delta, 2),
            ]
        )

        # Check if message is a 6065 protocol
        msg = speedwireHeader.from_packed(data[0:18])
        if not msg.check6065():
            _LOGGER.debug("Ignoring non 6065 Response. %d", msg.protokoll)
            return

        # If the requested information is not available, send the next command,
        if len(data) < 58:
            _LOGGER.debug(f"NACK [{len(data)}] -- {data!r}")
            self._confirm_repsonse()
            return

        # Handle Login Responses
        msg6065 = speedwireHeader6065.from_packed(data[18 : 18 + 36])
        if msg6065.isLoginResponse():
            self.handle_login(msg6065)
            self._confirm_repsonse()
            return

        # Filter out non matching responses
        (cnt_registers, size_registers) = self.calc_register(data, msg6065)
        code = int.from_bytes(data[54:58], "little")
        codem = code & 0x00FFFF00
        if len(data) == 58 and codem == 0:
            _LOGGER.debug(f"NACK [{len(data)}] -- {data!r}")
            self._confirm_repsonse()
            return
        if size_registers <= 0 or size_registers not in [16, 28, 40]:
            _LOGGER.warning(
                f"Skipping message. --- Len {data!r} Ril {codem} {cnt_registers} x {size_registers} bytes"
            )
            self._confirm_repsonse(code)
            return

        # Extract the values for each register
        for idx in range(0, cnt_registers):
            start = idx * size_registers + 54
            self.handle_register(data[start : start + size_registers], idx)

        self._confirm_repsonse(code)


class SMAspeedwireINV(Device):
    """Adapter between Device-Class and SMAClientProtocol"""

    _options: Dict[str, Any] = {}
    _transport = None
    _protocol = None
    _deviceinfo: DeviceInformation
    _debug: Dict[str, Any] = {"overalltimeout": 0}

    def __init__(self, host: str, group: str, password: Optional[str]):
        self._host = host
        self._group = group
        self._password = password
        if group not in ["user", "installer"]:
            raise KeyError(f"Invalid user type: {group} (user or installer)")

    async def _createEndpoint(self) -> None:
        if self._protocol is not None:
            _LOGGER.warning("Protocol already created")
            return
        loop = asyncio.get_running_loop()
        on_connection_lost = loop.create_future()
        if not self._password:
            raise ValueError("Password not set!")
        self._transport, self._protocol = await loop.create_datagram_endpoint(
            lambda: SMAClientProtocol(
                self._password,  # type: ignore[arg-type]
                on_connection_lost,
                self._options,
            ),
            remote_addr=(self._host, 9522),
        )

    # @override
    async def new_session(self) -> bool:
        # Create Endpoint
        await self._createEndpoint()
        if self._protocol is None:
            raise SmaConnectionException("protocol not initialized")

        self._protocol._failedCounter = 0
        self._protocol._sendCounter = 0
        # Test with device_info if the ip and user/pwd are correct
        await self.device_info()
        if self._protocol._failedCounter >= self._protocol._sendCounter:
            raise SmaConnectionException(
                f"No connection to device: {self._host}:9522  ({self._protocol._failedCounter}/{self._protocol._sendCounter})"
            )
        return True

    # @override
    async def device_info(self) -> dict:
        ll = await self.device_list()
        return list(ll.values())[0].asDict()

    # @override
    async def device_list(self) -> dict[str, DeviceInformation]:
        if self._protocol is None:
            raise SmaConnectionException("protocol not initialized")

        fut = asyncio.get_running_loop().create_future()
        await self._protocol.start_query(["TypeLabel", "Firmware"], fut, self._group)
        try:
            await asyncio.wait_for(fut, timeout=self._protocol._overallTimeout)
        except TimeoutError:
            self._debug["overalltimeout"] += 1
            _LOGGER.warning("Timeout in device_info")
            if (
                "error" in self._protocol.data_values
                and self._protocol.data_values["error"] == 0
            ):
                raise SmaReadException(
                    "Reply for request not received"
                )  # Recheck Logic
            raise SmaConnectionException("No connection to device")
        data = self._protocol.data_values

        invcnr = data.get("inverter_class", 0)
        invc = SMATagList.get(invcnr, f"Unknown device ({invcnr})")

        invtnr = data.get("inverter_type", 0)
        invt = SMATagList.get(invtnr, f"Unknown type ({invtnr})")

        self._deviceinfo = DeviceInformation(
            data.get("serial", ""),
            data.get("serial", ""),
            str(invt),
            str(invc),
            "SMA",
            data.get("Firmware", ""),
        )
        return {data.get("serial", ""): self._deviceinfo}

    # @override
    async def get_sensors(self, deviceID: str | None = None) -> Sensors:
        if self._protocol is None:
            raise SmaConnectionException("protocol not initialized")

        fut = asyncio.get_running_loop().create_future()
        c = self._protocol.allCmds
        device_sensors = Sensors()
        try:
            await self._protocol.start_query(c, fut, self._group)
            await asyncio.wait_for(fut, timeout=self._protocol._overallTimeout)
            for s in self._protocol.sensors.values():
                device_sensors.add(s)
        except asyncio.TimeoutError as e:
            self._debug["overalltimeout"] += 1
            raise e
        return device_sensors

    # @override
    async def read(self, sensors: Sensors, deviceID: str | None = None) -> bool:
        if self._protocol is None:
            raise SmaConnectionException("protocol not initialized")

        fut = asyncio.get_running_loop().create_future()
        c = self._protocol.allCmds
        await self._protocol.start_query(c, fut, self._group)
        try:
            await asyncio.wait_for(fut, timeout=self._protocol._overallTimeout)
            self._update_sensors(sensors, self._protocol.sensors, deviceID)
            return True
        except asyncio.TimeoutError as e:
            self._debug["overalltimeout"] += 1
            raise e

    def _update_sensors(
        self, sensors: Sensors, sensorReadings: dict[str, Sensor], deviceID: str | None
    ) -> None:
        """Update a sensor with the sensor reading"""
        _LOGGER.debug("Received %d sensor readings", len(sensorReadings))
        for sen in sensors:
            if sen.enabled and sen.key in sensorReadings:
                value = sensorReadings[sen.key].value
                if sen.mapper:
                    sen.mapped_value = sen.mapper.get(value, str(value))
                sen.value = value

    async def close_session(self) -> None:
        if self._transport is not None:
            self._transport.close()

    async def get_debug(self) -> Dict:
        if self._protocol is None:
            raise SmaConnectionException("protocol not initialized")

        ret = self._protocol.debug.copy()
        ret["unfinished"] = list(ret["unfinished"])
        ret["msg"] = list(ret["msg"])
        ret["ids"] = list(ret["ids"])
        ret["device_info"] = self._deviceinfo
        ret["timeouts"] = self._debug["overalltimeout"]
        return ret

    # wait for a response or a timeout
    async def detect(self, ip: str) -> list[DiscoveryInformation]:
        try:
            if self._protocol is None:
                raise SmaConnectionException("protocol not initialized")
            di = DiscoveryInformation()
            di.tested_endpoints = str(ip) + ":9522"
            await self.new_session()
            fut = asyncio.get_running_loop().create_future()
            await self._protocol.start_query(["TypeLabel"], fut, self._group)
            try:
                await asyncio.wait_for(fut, timeout=5)
            except TimeoutError:
                _LOGGER.warning("Timeout in detect")
            if (
                "error" in self._protocol.data_values
                and self._protocol.data_values["error"] == 0
            ):
                raise SmaReadException("Reply for request not received")
            raise SmaConnectionException("No connection to device")
        except SmaAuthenticationException as e:
            di.status = "maybe"
            di.exception = e
            di.remark = "only unencrypted Speedwire is supported"
        except Exception as e:
            di.status = "failed"
            di.exception = e
        return [di]

    def set_options(self, options: Dict[str, Any]) -> None:
        self._options = options
